#ifndef MSP_LINAL_VECTOR_H_
#define MSP_LINAL_VECTOR_H_

#include <algorithm>
#include <cmath>

namespace Msp {
namespace LinAl {

/**
Base class to provide the components of a vector.  This is used so that
specializations with individual members can be provided in some dimensions.
*/
template<typename T, unsigned N>
class VectorComponents
{
private:
	T data[N];

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return data[i]; }
	const T &operator[](unsigned i) const { return data[i]; }
};

template<typename T>
class VectorComponents<T, 2>
{
public:
	T x, y;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

template<typename T>
class VectorComponents<T, 3>
{
public:
	T x, y, z;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

template<typename T>
class VectorComponents<T, 4>
{
public:
	T x, y, z, w;

protected:
	VectorComponents() { }

public:
	T &operator[](unsigned i) { return *(&x+i); }
	const T &operator[](unsigned i) const { return *(&x+i); }
};

/**
A general mathematical vector.
*/
template<typename T, unsigned N>
class Vector: public VectorComponents<T, N>
{
public:
	Vector();
	Vector(const T *);
	Vector(T, T);
	Vector(T, T, T);
	Vector(T, T, T, T);
	template<typename U>
	Vector(const Vector<U, N> &);
	template<typename U>
	Vector(const Vector<U, N-1> &, U);
	template<typename U>
	explicit Vector(const Vector<U, N+1> &);

	Vector &operator*=(T);
	Vector &operator/=(T);
	Vector &operator+=(const Vector &);
	Vector &operator-=(const Vector &);

	T norm() const;
	Vector &normalize();
};


template<typename T, unsigned N>
inline Vector<T, N>::Vector()
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = T();
}

template<typename T, unsigned N>
inline Vector<T, N>::Vector(const T *d)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = d[i];
}

/* The compiler won't instantiate these unless they are used.  Trying to use
them on the wrong class results in an error. */
template<typename T, unsigned N>
inline Vector<T, N>::Vector(T x_, T y_)
{
	this->VectorComponents<T, 2>::x = x_;
	this->VectorComponents<T, 2>::y = y_;
}

template<typename T, unsigned N>
inline Vector<T, N>::Vector(T x_, T y_, T z_)
{
	this->VectorComponents<T, 3>::x = x_;
	this->VectorComponents<T, 3>::y = y_;
	this->VectorComponents<T, 3>::z = z_;
}

template<typename T, unsigned N>
inline Vector<T, N>::Vector(T x_, T y_, T z_, T w_)
{
	this->VectorComponents<T, 4>::x = x_;
	this->VectorComponents<T, 4>::y = y_;
	this->VectorComponents<T, 4>::z = z_;
	this->VectorComponents<T, 4>::w = w_;
}

template<typename T, unsigned N>
template<typename U>
inline Vector<T, N>::Vector(const Vector<U, N> &v)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = v[i];
}

template<typename T, unsigned N>
template<typename U>
inline Vector<T, N>::Vector(const Vector<U, N-1> &v, U s)
{
	for(unsigned i=0; i<N-1; ++i)
		(*this)[i] = v[i];
	(*this)[N-1] = s;
}

template<typename T, unsigned N>
template<typename U>
inline Vector<T, N>::Vector(const Vector<U, N+1> &v)
{
	for(unsigned i=0; i<N; ++i)
		(*this)[i] = v[i];
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator*=(T s)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] *= s;
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator*(const Vector<T, N> &v, T s)
{
	Vector<T, N> r(v);
	return r *= s;
}

template<typename T, unsigned N>
inline Vector<T, N> operator*(T s, const Vector<T, N> &v)
{
	return v*s;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator/=(T s)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] /= s;
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator/(const Vector<T, N> &v, T s)
{
	Vector<T, N> r(v);
	return r /= s;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator+=(const Vector<T, N> &v)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] += v[i];
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator+(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	Vector<T, N> r(v1);
	return r += v2;
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::operator-=(const Vector<T, N> &v)
{
	for(unsigned i=0; i<N; ++i)	
		(*this)[i] -= v[i];
	return *this;
}

template<typename T, unsigned N>
inline Vector<T, N> operator-(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	Vector<T, N> r(v1);
	return r -= v2;
}

template<typename T, unsigned N>
inline Vector<T, N> operator-(const Vector<T, N> &v)
{
	Vector<T, N> r(v);
	for(unsigned i=0; i<N; ++i)
		r[i] = -r[i];
	return r;
}

template<typename T, unsigned N>
inline bool operator==(const Vector<T, N> &v, const Vector<T, N> &w)
{
	for(unsigned i=0; i<N; ++i)
		if(v[i]!=w[i])
			return false;
	return true;
}

template<typename T, unsigned N>
inline T inner_product(const Vector<T, N> &v1, const Vector<T, N> &v2)
{
	T r = T();
	for(unsigned i=0; i<N; ++i)
		r += v1[i]*v2[i];
	return r;
}

template<typename T, unsigned N>
inline T Vector<T, N>::norm() const
{
	using std::sqrt;
	return sqrt(inner_product(*this, *this));
}

template<typename T, unsigned N>
inline Vector<T, N> &Vector<T, N>::normalize()
{
	return *this /= norm();
}

template<typename T, unsigned N>
inline Vector<T, N> normalize(const Vector<T, N> &v)
{
	Vector<T, N> r(v);
	return r.normalize();
}

template<typename T>
inline T dot(const Vector<T, 3> &v1, const Vector<T, 3> &v2)
{
	return inner_product(v1, v2);
}

template<typename T>
inline Vector<T, 3> cross(const Vector<T, 3> &v1, const Vector<T, 3> &v2)
{
	return Vector<T, 3>(v1.y*v2.z-v1.z*v2.y, v1.z*v2.x-v1.x*v2.z, v1.x*v2.y-v1.y*v2.x);
}

} // namespace LinAl
} // namespace Msp

#endif
